#
#  PCC Debugger Module
#

sub NRESULTS { 100; }           # Console::nResults
sub OBJS_PER_PD { 1000; }       # Swapper::OBJS_PER_PD
sub HASH_VALUE { 64; }          # ccExec::HASH_VALUE
sub SIZEOF_SHIPXY { 30; }       # Sizeof(TShipXY)

sub get_string {
    my $a = shift;
    my $seg = ($a >> 16) & 0xFFFF;
    my $ofs = $a & 0xFFFF;
    my $len = ord &get_memory ($seg, $ofs, 1);

    &get_memory ($seg, $ofs+1, $len);
};

#
#  COMMAND: console
#
#  Shows the last output on the PCC console. With numerical argument,
#  show at most that many lines. When invoked with a regexp as argument,
#  such as `console/PCC', show only lines matching that re. The regexp
#  can be negated (`console !/PCC').
#
$command_map{'console'} = sub {
    my $amount = shift || -1;
    my $re = undef;
    my $reinverse = 0;
    if($amount =~ m,^(\!?/)(.*),) {
        $re = $2;
        $reinverse = ($1 ne "/");
        $amount = shift || -1;
    }

    my $consoleptrs = &get_sym_val("RESULTS", 4*&NRESULTS);
    if(!defined $consoleptrs) {
        print STDERR "Unable to find console data in image\n";
        return;
    }
    my @consoleptrs = unpack "V".(&NRESULTS), $consoleptrs;
    my @results = ();
    foreach(@consoleptrs) {
        next if !$_;
        my $seg = ($_ >> 16) & 0xFFFF;
        my $ofs = $_ & 0xFFFF;
        my $len = ord &get_memory($seg, $ofs+2, 1);
        my $str = &get_memory($seg, $ofs+3, $len);
        my $just = ord &get_memory($seg, $ofs, 1);
        my $color = ord &get_memory($seg, $ofs+1, 1);
        if(defined $re) {
            next if $reinverse && $str =~ /$re/;
            next if !$reinverse && $str !~ /$re/;
        }
        if ($just || $color) {
            my $pfx = "";
            $pfx .= ("<" . ($color & 127) . ">") if ($color & 127);
            $pfx .= "! " if ($color & 128);
            $pfx .= ">  " if ($just == 2);
            $str = $pfx . $str;
        }
        push @results, $str;
    }
    if($amount >= 0) {
        shift @results while $amount <= $#results;
    }
    print join "\n", @results, "";
};
$help .= "  console [n]    show last [n lines of] console output\n";

#
#  COMMAND: swap
#
#  Shows the swapper status, that is, all swapper-controlled objects.
#  With `more' function
#
$command_map{'swap'} = sub {
    my $pd = &get_sym_val ("FIRST_PD", 4);
    if (!defined $pd) {
        print STDERR "Unable to find swapper page directory in image\n";
        return;
    }
    my $count = 0;
    while ($pd = (unpack "V", $pd)[0]) {
        my $dir = &get_memory (($pd >> 16) & 0xFFFF, $pd & 0xFFFF, 4 + &OBJS_PER_PD * 14);
        foreach (0..(&OBJS_PER_PD - 1)) {
            my $entry = substr ($dir, 4 + 14*$_);
            my @entry = unpack "ccVVvv", $entry;
            next if $entry[0] == 0;

            my $status = "";
            $status .= 'S' if $entry[0] & 128;
            $status .= 'P' if $entry[0] & 64;
            $status .= 's' if $entry[0] & 8;
            $status .= 'r' if $entry[0] & 4;
            $status .= ('-', '<data>', '<res>', '<cres>')[$entry[0] & 3];

            if ($count == 23) {
                print "--More (q=exit)--";
                return if <> =~ /q/;
                $count=0;
            }
                  # 123:1234567890 123 12345678 1234567890 12345 12345
            print  "Status________ Lck LoadAdr_ SwapAdr___ Size_ LRU__\n" if $count==0;
            printf "%3d:%-10s %3d %08x %10d %5d %5d\n", $entry[0], $status,
                $entry[1], $entry[2], $entry[3], $entry[4], $entry[5];
            $count++;
        }
        $pd = substr ($dir, 0, 4);
    }
};
$help .= "  swap           show swapper status\n";

#
#  COMMAND: hash
#
#  Show subr_hash ("script object registry")
#
$command_map{'hash'} = sub {
#            code     : PObj;           { Code als Folge von Strings }
#            min, max : INTEGER;        { Argumentzahlen }
#            argnames : CNameList;      { Argumentnamen }
#            name     : PString;
#            next     : PSubroutine;
#            ref      : WORD;           { Referenzzhler }
#            slot     : BYTE;           { Backlink subr_hash-Slot }
#            typ      : BYTE;
    $subrhash = &get_sym_val ("SUBR_HASH", 4 * &HASH_VALUE);
    if (!defined $subrhash) {
        print STDERR "Unable to find subr_hash in image\n";
        return;
    }
    my $count = 0;
    foreach (unpack "V".(&HASH_VALUE), $subrhash) {
        my $seg = ($_ >> 16) & 0xFFFF;
        my $ofs = $_ & 0xFFFF;
        while ($seg) {
            my ($code, $min, $max, $argnames, $name, $next, $ref, $slot, $typ)
                = unpack "VvvA8VVvcc", &get_memory ($seg, $ofs, 28);
            if ($name) {
                $name = &get_string ($name);
            } else {
                $name = "(null)";
            }
            if ($count == 23) {
                print "--More (q=exit)--";
                return if <> =~ /q/;
                $count=0;
            }
            my $s = sprintf "%s (%d..%d)", $name, $min, $max;
                  # 1234:1234  123456789012345678901234567890 123 1234567 12345678 123
            print  "Subr_____  Name__________________________ Ref Type___ Code____ Lck\n" if $count==0;
            printf "%04x:%04x: %-30s %3d %-7s %08x %3d\n",
                $seg, $ofs, $s, $ref,
                ('code', 'keymap')[$typ],
                $code,
                ord (&get_memory(($code >> 16) & 0xFFFF, ($code & 0xFFFF) + 1, 1));
            $seg = ($next >> 16) & 0xFFFF;
            $ofs = $next & 0xFFFF;
            $count++;
        }
    }
};
$help .= "  hash           show subr_hash\n";

#
#  COMMAND: fleet [id]
#
#  Shows fleets (just fleet #id)
#
$command_map{'fleet'} = sub {
    my $arg = shift;
    my $ptr = get_sym_val ("FLEETS", 4);
    my $sptr = get_sym_val ("SHIPS", 4);
    if (!defined $ptr) { print STDERR "Unable to find fleet pointer in image\n"; return; }
    if (!defined $sptr) { print STDERR "Unable to find ship pointer in image\n"; return; }
    my $count = 0;
    my ($ofs,$seg) = unpack "vv", $ptr;
    my ($sofs,$sseg) = unpack "vv", $sptr;
    if ($seg == 0 && $ofs == 0) {
        print STDERR "Fleet pointer is NIL\n";
        return;
    }
    if ($sofs==0 && $sseg==0) {
        print "Note: ship data not accessible\n";
        ++$count;
    }
    foreach (1..999) {
        my $entry = get_memory ($seg, $ofs + 6*($_-1), 6);
        if(!defined $entry) {
            print "End of segment reached at Id $_\n";
            last;
        }
        my ($id, $nameptr) = unpack "vV", $entry;
        next if $id==0 && $nameptr==0;
        my $name = '';
        if ($nameptr) { $name = ', Name: "' . get_string($nameptr) . '"'; }
        if (!defined $arg || $arg == $id || $arg == $_) {
            my $status = ' (no ship data)';
            if ($sofs!=0 || $sseg!=0) {
                my $sdata = get_memory($sseg, $sofs + SIZEOF_SHIPXY() * ($_-1), SIZEOF_SHIPXY());
                if (defined $sdata) {
                    my ($x,$y,$own,$mass,$orbit,$dat,$huser,$target,
                        $hhist,$rc,$text,$st) = unpack "vvvvvvVVVcVc", $sdata;
                    $status = "  player-$own";
                    $status .= " at-$x,$y" if $x || $y;
                    $status .= " dat" if $huser;
                    $status .= " target" if $target;
                    $status .= " history" if $hhist;
                } else {
                    $status = " (ship data out of range)";
                }
            }

            printf "Ship %3d:  Fleet %3d%s%s\n", $_, $id, $name, $status;
            if (++$count == 23) {
                print "--More (q=exit)--";
                return if <> =~ /q/;
                $count=0;
            }
        }
    }
};
$help .= "  fleet [n]      show fleets (only fleet n)\n";

print "PCC debugger module loaded.\n";
1;
